<?php

namespace apexl\Io\modules\system\services;

use apexl\ClassEngine\EngineSingleton;
use apexl\Config\Configuration;
use apexl\Vault\Vault;

class cache extends EngineSingleton
{
    /** @var Singleton */
    protected $config;

    /** @var Vault */
    protected $database;

    protected function __construct()
    {
        parent::__construct();
        $this->config = Configuration::getInstance();
    }
}
